import torchbearer

from torchbearer.callbacks import Callback


class EarlyStopping(Callback):
    """Callback to stop training when a monitored quantity has stopped improving.
    """

    def __init__(self, monitor='val_loss',
                 min_delta=0, patience=0, verbose=0, mode='auto'):
        """Stop training when a monitored quantity has stopped improving.

        :param monitor: Quantity to be monitored
        :type monitor: str
        :param min_delta: Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute change
        of less than min_delta, will count as no improvement.
        :type min_delta: float
        :param patience: Number of epochs with no improvement after which training will be stopped.
        :type patience: int
        :param verbose: Verbosity mode, will print stopping info if verbose > 0
        :type verbose: int
        :param mode: One of {auto, min, max}. In `min` mode, training will stop when the quantity monitored has stopped
        decreasing; in `max` mode it will stop when the quantity monitored has stopped increasing; in `auto` mode, the
        direction is automatically inferred from the name of the monitored quantity.
        :type mode: str
        """
        super(EarlyStopping, self).__init__()

        self.monitor = monitor
        self.min_delta = min_delta
        self.patience = patience
        self.verbose = verbose
        self.mode = mode

        self.wait = 0
        self.stopped_epoch = 0

        if self.mode not in ['min', 'max']:
            if 'acc' in self.monitor:
                self.mode = 'max'
            else:
                self.mode = 'min'

        if self.mode == 'min':
            self.min_delta *= -1
            self.monitor_op = lambda x1, x2: x1 < x2
        elif self.mode == 'max':
            self.min_delta *= 1
            self.monitor_op = lambda x1, x2: x1 > x2

    def on_start(self, state):
        self.wait = 0
        self.stopped_epoch = 0
        self.best = float('inf') if self.mode == 'min' else -float('inf')

    def on_end_epoch(self, state):
        current = state[torchbearer.METRICS][self.monitor]
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = state[torchbearer.EPOCH]
                state[torchbearer.STOP_TRAINING] = True

    def on_end(self, state):
        if self.stopped_epoch > 0 and self.verbose > 0:
            print('Epoch %05d: early stopping' % (self.stopped_epoch + 1))



import unittest

import torchbearer


class TestStateKey(unittest.TestCase):
    def test_key_added(self):
        key = torchbearer.state_key('key')

        self.assertTrue(key in torchbearer.STATE_KEYS)

    def test_duplicate(self):
        key = torchbearer.state_key(torchbearer.MODEL)

        self.assertTrue(torchbearer.MODEL != key)



import torchbearer

from torchbearer.callbacks import Callback

import torch


class WeightDecay(Callback):
    """Callback which adds a weight decay term to the loss for the given parameters.
    """

    def __init__(self, rate=5e-4, p=2, params=None):
        """Create a WeightDecay callback which uses the given norm on the given parameters and with the given decay
        rate. If params is None (default) then the parameters will be retrieved from the model.

        :param rate: The decay rate
        :type rate: float
        :param p: The norm level
        :type p: int
        :param params: The parameters to use (or None)
        :type params: list
        """
        super(WeightDecay, self).__init__()

        self.p = p
        self.params = params
        self.rate = rate / self.p

    def on_start(self, state):
        """Retrieve params from state['model'] if required.

        :param state: The Model state
        :type state: dict
        """
        if self.params is None:
            self.params = state[torchbearer.MODEL].parameters()

    def on_criterion(self, state):
        """Calculate the decay term and add to state['loss'].

        :param state: The Model state
        :type state: dict
        """
        for param in self.params:
            state['loss'] += self.rate * torch.pow(param, self.p).sum()


class L1WeightDecay(WeightDecay):
    """WeightDecay callback which uses an L1 norm
    """
    def __init__(self, rate=5e-4, params=None):
        """Create L1WeightDecayCallback with the given rate and parameters. If params is None (default) then the
        parameters will be retrieved from the model.

        :param rate: The decay rate
        :type rate: float
        :param params: The parameters to use (or None)
        :type params: list
        """
        super(L1WeightDecay, self).__init__(rate=rate, p=1, params=params)


class L2WeightDecay(WeightDecay):
    """WeightDecay callback which uses an L2 norm
    """
    def __init__(self, rate=5e-4, params=None):
        """Create L2WeightDecayCallback with the given rate and parameters. If params is None (default) then the
        parameters will be retrieved from the model.

        :param rate: The decay rate
        :type rate: float
        :param params: The parameters to use (or None)
        :type params: list
        """
        super(L2WeightDecay, self).__init__(rate=rate, p=2, params=params)



"""
Base metrics are the base classes which represent the metrics supplied with torchbearer. The all use the
:func:`.default_for_key` decorator so that they can be accessed in the call to :class:`.torchbearer.Model` via the
following strings:

- '`acc`' or '`accuracy`': The :class:`.CategoricalAccuracy` metric
- '`loss`': The :class:`.Loss` metric
- '`epoch`': The :class:`.Epoch` metric
- '`roc_auc`' or '`roc_auc_score`': The :class:`.RocAucScore` metric
"""
import torchbearer
from torchbearer import metrics

import torch


class CategoricalAccuracy(metrics.BatchLambda):
    """Categorical accuracy metric. Uses torch.max to determine predictions and compares to targets.
    """

    def __init__(self):
        def metric_function(y_pred, y_true):
            _, y_pred = torch.max(y_pred, 1)
            return (y_pred == y_true).float()
        super(CategoricalAccuracy, self).__init__('acc', metric_function)


@metrics.default_for_key('acc')
@metrics.default_for_key('accuracy')
@metrics.running_mean
@metrics.std
@metrics.mean
class CategoricalAccuracyFactory(metrics.MetricFactory):
    def build(self):
        return CategoricalAccuracy()


class Loss(metrics.Metric):
    """Simply returns the 'loss' value from the model state.
    """

    def __init__(self):
        super().__init__('loss')

    def process(self, state):
        return state[torchbearer.LOSS]


@metrics.default_for_key('loss')
@metrics.running_mean
@metrics.std
@metrics.mean
class LossFactory(metrics.MetricFactory):
    def build(self):
        return Loss()


class Epoch(metrics.Metric):
    """Returns the 'epoch' from the model state.
    """

    def __init__(self):
        super().__init__('epoch')

    def process_final(self, state):
        return self._process(state)

    def process(self, state):
        return self._process(state)

    def _process(self, state):
        return state[torchbearer.EPOCH]


@metrics.default_for_key('epoch')
@metrics.to_dict
class EpochFactory(metrics.MetricFactory):
    def build(self):
        return Epoch()



import unittest
from unittest.mock import patch, Mock

import torchbearer.metrics as metrics
from torchbearer.metrics import default_for_key, lambda_metric, EpochLambda, BatchLambda


class TestDecorators(unittest.TestCase):
    def test_default_for_key_class(self):
        metric = metrics.Loss
        metric = default_for_key('test')(metric)
        self.assertTrue('test' in metrics.DEFAULT_METRICS)
        self.assertTrue(metrics.DEFAULT_METRICS['test'].name == 'loss')
        self.assertTrue(metric == metrics.Loss)

    def test_default_for_key_metric(self):
        metric = metrics.Loss()
        metric = default_for_key('test')(metric)
        self.assertTrue('test' in metrics.DEFAULT_METRICS)
        self.assertTrue(metrics.DEFAULT_METRICS['test'].name == 'loss')
        self.assertTrue(metric.name == 'loss')

    def test_lambda_metric_epoch(self):
        metric = 'test'
        metric = lambda_metric('test', on_epoch=True)(metric)().build()
        self.assertTrue(isinstance(metric, EpochLambda))
        self.assertTrue(metric._final == 'test')

    def test_lambda_metric_batch(self):
        metric = 'test'
        metric = lambda_metric('test')(metric)().build()
        self.assertTrue(isinstance(metric, BatchLambda))
        self.assertTrue(metric._metric_function == 'test')

    def test_to_dict_metric(self):
        metric = metrics.Metric
        out = metrics.to_dict(metric)('test').build()
        self.assertTrue(out.metric.name == 'test')
        self.assertTrue(isinstance(out, metrics.ToDict))

    @patch('torchbearer.metrics.MetricFactory.build')
    def test_to_dict_metric_factory(self, build_mock):
        metric = metrics.MetricFactory
        build_mock.return_value = metrics.Metric('test')
        out = metrics.to_dict(metric)().build()
        self.assertTrue(out.metric.name == 'test')
        self.assertTrue(isinstance(out, metrics.ToDict))
        build_mock.assert_called_once()

    def test_mean_metric(self):
        metric = metrics.Metric
        out = metrics.mean(metric)('test').build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Mean))
        self.assertTrue(out.children[0].metric.name == 'test')
        self.assertTrue(out.root.name == 'test')

    @patch('torchbearer.metrics.MetricFactory.build')
    def test_mean_metric_factory(self, build_mock):
        metric = metrics.MetricFactory
        build_mock.return_value = metrics.Metric('test')
        out = metrics.mean(metric)().build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Mean))
        self.assertTrue(out.children[0].metric.name == 'test')
        self.assertTrue(out.root.name == 'test')

    def test_std_metric(self):
        metric = metrics.Metric
        out = metrics.std(metric)('test').build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Std))
        self.assertTrue(out.children[0].metric.name == 'test_std')
        self.assertTrue(out.root.name == 'test')

    @patch('torchbearer.metrics.MetricFactory.build')
    def test_std_metric_factory(self, build_mock):
        metric = metrics.MetricFactory
        build_mock.return_value=metrics.Metric('test')
        out = metrics.std(metric)().build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Std))
        self.assertTrue(out.children[0].metric.name == 'test_std')
        self.assertTrue(out.root.name == 'test')

    def test_running_mean_metric(self):
        metric = metrics.Metric
        out = metrics.running_mean(batch_size=40, step_size=20)(metric)('test').build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.RunningMean))
        self.assertTrue(out.children[0].metric._batch_size == 40)
        self.assertTrue(out.children[0].metric._step_size == 20)
        self.assertTrue(out.children[0].metric.name == 'running_test')
        self.assertTrue(out.root.name == 'test')

    @patch('torchbearer.metrics.MetricFactory.build')
    def test_running_mean_metric_factory(self, build_mock):
        metric = metrics.MetricFactory
        build_mock.return_value=metrics.Metric('test')
        out = metrics.running_mean(metric)().build()
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.RunningMean))
        self.assertTrue(out.children[0].metric.name == 'running_test')
        self.assertTrue(out.root.name == 'test')



from unittest import TestCase
from unittest.mock import patch, Mock

from torchbearer.callbacks import TimerCallback


class TestTimer(TestCase):
    @patch('time.time')
    def test_update_time(self, time):
        time.return_value = 0
        timer = TimerCallback()
        time.return_value = 1
        timer.update_time('test', {})
        self.assertTrue(timer.get_timings()['test'] == 1)

        time.return_value = 3
        timer.update_time('test_2', {})
        self.assertTrue(timer.get_timings()['test_2'] == 2)

    def test_calls(self):
        timer = TimerCallback()
        timer.update_time = Mock()

        timer.on_start({})
        timer.on_start_training({})
        timer.on_start_epoch({})
        timer.on_sample({})
        timer.on_forward({})
        timer.on_criterion({})
        timer.on_backward({})
        timer.on_step_training({})
        timer.on_start_validation({})
        timer.on_sample_validation({})
        timer.on_forward_validation({})
        timer.on_criterion_validation({})
        timer.on_step_validation({})
        self.assertTrue(timer.update_time.call_count == 13)



"""
The base metric classes exist to enable complex data flow requirements between metrics. All metrics are either instances
of :class:`Metric` or :class:`MetricFactory`. These can then be collected in a :class:`MetricList` or a
:class:`MetricTree`. The :class:`MetricList` simply aggregates calls from a list of metrics, whereas the
:class:`MetricTree` will pass data from its root metric to each child and collect the outputs. This enables complex
running metrics and statistics, without needing to compute the underlying values more than once. Typically,
constructions of this kind should be handled using the :mod:`decorator API <.metrics.decorators>`.
"""

import abc

"""The global dict of default metrics which maps keys to metrics in the :class:`MetricList`.
"""
DEFAULT_METRICS = {}


class Metric(object):
    """Base metric class. Process will be called on each batch, process-final at the end of each epoch.
    The metric contract allows for metrics to take any args but not kwargs. The initial metric call will be given state,
    however, subsequent metrics can pass any values desired.

    .. note::

        All metrics must extend this class.

    :param name: The name of the metric
    :type name: str

    """

    def __init__(self, name):
        self.name = name

    def process(self, *args):
        """Process the state and update the metric for one iteration.

        :param args: Arguments given to the metric. If this is a root level metric, will be given state
        :return: None, or the value of the metric for this batch

        """
        ...

    def process_final(self, *args):
        """Process the terminal state and output the final value of the metric.

        :param args: Arguments given to the metric. If this is a root level metric, will be given state
        :return: None or the value of the metric for this epoch

        """
        ...

    def eval(self):
        """Put the metric in eval mode during model validation.
        """
        ...

    def train(self):
        """Put the metric in train mode during model training.
        """
        ...

    def reset(self, state):
        """Reset the metric, called before the start of an epoch.

        :param state: The current state dict of the :class:`.Model`.

        """
        ...


class MetricFactory(object):
    """A simple implementation of a factory pattern. Used to enable construction of complex metrics using decorators.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def build(self):
        """Build and return a usable :class:`Metric` instance.

        :return: The constructed :class:`Metric`
        """
        ...


class MetricTree(Metric):
    """A tree structure which has a node :class:`Metric` and some children. Upon execution, the node is called with the
    input and its output is passed to each of the children. A dict is updated with the results.

    :param metric: The metric to act as the root node of the tree / subtree
    :type metric: Metric
    """
    def __init__(self, metric):
        super(MetricTree, self).__init__(metric.name)
        self.root = metric
        self.children = []

    def add_child(self, child):
        """Add a child to this node of the tree

        :param child: The child to add
        :type child: Metric
        :return: None
        """
        self.children.append(child)

    def _for_tree(self, function, *args):
        result = {}
        node_value = function(self.root, *args)

        for subtree in self.children:
            out = function(subtree, node_value)
            if out is not None:
                result.update(out)

        return result

    def process(self, *args):
        """Process this node and then pass the output to each child.

        :return: A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process(*in_args), *args)

    def process_final(self, *args):
        """Process this node and then pass the output to each child.

        :return: A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process_final(*in_args), *args)

    def eval(self):
        self.root.eval()

        for subtree in self.children:
            subtree.eval()

    def train(self):
        self.root.train()

        for subtree in self.children:
            subtree.train()

    def reset(self, state):
        self.root.reset(state)

        for subtree in self.children:
            subtree.reset(state)


class MetricList(Metric):
    """The :class:`MetricList` class is a wrapper for a list of metrics which acts as a single metric and produces a
    dictionary of outputs.

    :param metric_list: The list of metrics to be wrapped. If the list contains a :class:`MetricList`, this will be\
    unwrapped. Any strings in the list will be retrieved from metrics.DEFAULT_METRICS.
    :type metric_list: list
    """

    def __init__(self, metric_list):
        super().__init__('metric_list')

        self.metric_list = []

        for i in range(len(metric_list)):
            metric = metric_list[i]

            if str(metric) == metric:
                metric = DEFAULT_METRICS[metric]

            if isinstance(metric, MetricFactory):
                metric = metric.build()

            if type(metric) is MetricList:
                self.metric_list = self.metric_list + metric.metric_list
            else:
                self.metric_list.append(metric)

    def _for_list(self, function):
        result = {}
        for metric in self.metric_list:
            out = function(metric)
            if out is not None:
                result.update(out)
        return result

    def process(self, state):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        :param state: The current state dict of the :class:`.Model`.
        :return: dict[str,any] -- A dictionary which maps metric names to values.

        """
        return self._for_list(lambda metric: metric.process(state))

    def process_final(self, state):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        :param state: The current state dict of the :class:`.Model`.
        :return: dict[str,any] -- A dictionary which maps metric names to values.

        """
        return self._for_list(lambda metric: metric.process_final(state))

    def train(self):
        """Put each metric in train mode.
        """
        self._for_list(lambda metric: metric.train())

    def eval(self):
        """Put each metric in eval mode
        """
        self._for_list(lambda metric: metric.eval())

    def reset(self, state):
        """Reset each metric with the given state.

        :param state: The current state dict of the :class:`.Model`.

        """
        self._for_list(lambda metric: metric.reset(state))


class AdvancedMetric(Metric):
    """The :class:`AdvancedMetric` class is a metric which provides different process methods for training and
    validation. This enables running metrics which do not output intermediate steps during validation.

    :param name: The name of the metric.
    :type name: str

    """

    def __init__(self, name):
        super(AdvancedMetric, self).__init__(name)
        self._process = self.process_train
        self._process_final = self.process_final_train

    def process_train(self, *args):
        """Process the given state and return the metric value for a training iteration.

        :param state: The current state dict of the :class:`.Model`.
        :return: The metric value for a training iteration.

        """
        ...

    def process_validate(self, *args):
        """Process the given state and return the metric value for a validation iteration.

        :param state: The current state dict of the :class:`.Model`.
        :return: The metric value for a validation iteration.

        """
        ...

    def process_final_train(self, *args):
        """Process the given state and return the final metric value for a training iteration.

        :param state: The current state dict of the :class:`.Model`.
        :return: The final metric value for a training iteration.

        """
        ...

    def process_final_validate(self, *args):
        """Process the given state and return the final metric value for a validation iteration.

        :param state: The current state dict of the :class:`.Model`.
        :type state: dict
        :return: The final metric value for a validation iteration.

        """
        ...

    def process(self, *args):
        """Depending on the current mode, return the result of either 'process_train' or 'process_validate'.

        :param state: The current state dict of the :class:`.Model`.
        :type state: dict
        :return: The metric value.

        """
        return self._process(*args)

    def process_final(self, *args):
        """Depending on the current mode, return the result of either 'process_final_train' or 'process_final_validate'.

        :param state: The current state dict of the :class:`.Model`.
        :type state: dict
        :return: The final metric value.

        """
        return self._process_final(*args)

    def eval(self):
        """Put the metric in eval mode.
        """
        self._process = self.process_validate
        self._process_final = self.process_final_validate

    def train(self):
        """Put the metric in train mode.
        """
        self._process = self.process_train
        self._process_final = self.process_final_train



from unittest import TestCase
from unittest.mock import patch, Mock

import torchbearer
from torchbearer.callbacks import Tqdm, ConsolePrinter


class TestConsolePrinter(TestCase):
    @patch('builtins.print')
    def test_console_printer(self, mock_print):
        state = {torchbearer.BATCH: 5, torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.TRAIN_STEPS: 100, torchbearer.VALIDATION_STEPS: 101, torchbearer.METRICS: {'test': 0.9945}}
        printer = ConsolePrinter(validation_label_letter='e')

        printer.on_step_training(state)
        mock_print.assert_called_once_with('\r1/10(t): 5/100 test:0.995', end='')
        mock_print.reset_mock()

        printer.on_end_training(state)
        mock_print.assert_called_once_with('\r1/10(t): test:0.995')
        mock_print.reset_mock()

        printer.on_step_validation(state)
        mock_print.assert_called_once_with('\r1/10(e): 5/101 test:0.995', end='')
        mock_print.reset_mock()

        printer.on_end_validation(state)
        mock_print.assert_called_once_with('\r1/10(e): test:0.995')
        mock_print.reset_mock()


class TestTqdm(TestCase):
    @patch("torchbearer.callbacks.printer.tqdm")
    def test_tqdm(self, mock_tqdm):
        mock_tqdm.return_value = Mock()
        mock_tqdm.return_value.set_postfix = Mock()
        mock_tqdm.return_value.close = Mock()
        mock_tqdm.return_value.update = Mock()

        state = {torchbearer.EPOCH: 1, torchbearer.MAX_EPOCHS: 10, torchbearer.TRAIN_STEPS: 100, torchbearer.VALIDATION_STEPS: 101, torchbearer.METRICS: 'test'}
        tqdm = Tqdm(validation_label_letter='e')

        tqdm.on_start_training(state)
        mock_tqdm.assert_called_once_with(total=100, desc='1/10(t)')

        tqdm.on_step_training(state)
        mock_tqdm.return_value.set_postfix.assert_called_once_with('test')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix.reset_mock()

        tqdm.on_end_training(state)
        mock_tqdm.return_value.set_postfix.assert_called_once_with('test')
        mock_tqdm.return_value.close.assert_called_once()

        mock_tqdm.reset_mock()
        mock_tqdm.return_value.set_postfix.reset_mock()
        mock_tqdm.return_value.update.reset_mock()
        mock_tqdm.return_value.close.reset_mock()

        tqdm.on_start_validation(state)
        mock_tqdm.assert_called_once_with(total=101, desc='1/10(e)')

        tqdm.on_step_validation(state)
        mock_tqdm.return_value.set_postfix.assert_called_once_with('test')
        mock_tqdm.return_value.update.assert_called_once_with(1)
        mock_tqdm.return_value.set_postfix.reset_mock()

        tqdm.on_end_validation(state)
        mock_tqdm.return_value.set_postfix.assert_called_once_with('test')
        mock_tqdm.return_value.close.assert_called_once()



class Callback(object):
    """Base callback class.

    .. note::

        All callbacks should override this class.

    """

    def on_start(self, state):
        """Perform some action with the given state as context at the start of a model fit.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_start_epoch(self, state):
        """Perform some action with the given state as context at the start of each epoch.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_start_training(self, state):
        """Perform some action with the given state as context at the start of the training loop.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_sample(self, state):
        """Perform some action with the given state as context after data has been sampled from the generator.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_forward(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_criterion(self, state):
        """Perform some action with the given state as context after the criterion has been evaluated.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_backward(self, state):
        """Perform some action with the given state as context after backward has been called on the loss.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_step_training(self, state):
        """Perform some action with the given state as context after step has been called on the optimiser.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_end_training(self, state):
        """Perform some action with the given state as context after the training loop has completed.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_end_epoch(self, state):
        """Perform some action with the given state as context at the end of each epoch.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_end(self, state):
        """Perform some action with the given state as context at the end of the model fitting.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_start_validation(self, state):
        """Perform some action with the given state as context at the start of the validation loop.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_sample_validation(self, state):
        """Perform some action with the given state as context after data has been sampled from the validation generator.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_forward_validation(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed
        with the validation data.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_criterion_validation(self, state):
        """Perform some action with the given state as context after the criterion evaluation has been completed
        with the validation data.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_end_validation(self, state):
        """Perform some action with the given state as context at the end of the validation loop.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass

    def on_step_validation(self, state):
        """Perform some action with the given state as context at the end of each validation step.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        pass


class CallbackList(Callback):
    """The :class:`CallbackList` class is a wrapper for a list of callbacks which acts as a single callback.
    """

    def __init__(self, callback_list):
        """Create a new callback which wraps and internally calls each callback in the given list in turn.

        :param callback_list:The list of callbacks to be wrapped. If the list contains a :class:`CallbackList`, this
        will be unwrapped.
        :type callback_list:list

        """
        super().__init__()
        self.callback_list = []

        for i in range(len(callback_list)):
            callback = callback_list[i]

            if type(callback) is CallbackList:
                self.callback_list = self.callback_list + callback.callback_list
            else:
                self.callback_list.append(callback)

    def _for_list(self, function):
        for callback in self.callback_list:
            function(callback)
        
    def on_start(self, state):
        """Call on_start on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_start(state))

    def on_start_epoch(self, state):
        """Call on_start_epoch on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_start_epoch(state))

    def on_start_training(self, state):
        """Call on_start_training on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_start_training(state))

    def on_sample(self, state):
        """Call on_sample on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_sample(state))

    def on_forward(self, state):
        """Call on_forward on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_forward(state))

    def on_criterion(self, state):
        """Call on_criterion on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_criterion(state))

    def on_backward(self, state):
        """Call on_backward on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_backward(state))

    def on_step_training(self, state):
        """Call on_step_training on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_step_training(state))

    def on_end_training(self, state):
        """Call on_end_training on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_end_training(state))

    def on_end_epoch(self, state):
        """Call on_end_epoch on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_end_epoch(state))

    def on_end(self, state):
        """Call on_end on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_end(state))

    def on_start_validation(self, state):
        """Call on_start_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_start_validation(state))

    def on_sample_validation(self, state):
        """Call on_sample_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_sample_validation(state))

    def on_forward_validation(self, state):
        """Call on_forward_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_forward_validation(state))

    def on_criterion_validation(self, state):
        """Call on_criterion_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_criterion_validation(state))

    def on_end_validation(self, state):
        """Call on_end_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_end_validation(state))

    def on_step_validation(self, state):
        """Call on_step_validation on each callback in turn with the given state.

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]

        """
        self._for_list(lambda callback: callback.on_step_validation(state))



import copy

import torchvision.utils as utils

from tensorboardX import SummaryWriter

import torch.nn.functional as F

import torchbearer

from torchbearer.callbacks import Callback

import os

import torch


class TensorBoard(Callback):
    """The TensorBoard callback is used to write metric graphs to tensorboard. Requires the TensorboardX library for
    python.
    """

    def __init__(self, log_dir='./logs',
                 write_graph=True,
                 write_batch_metrics=False,
                 batch_step_size=10,
                 write_epoch_metrics=True,
                 comment='torchbearer'):
        """TensorBoard callback which writes metrics to the given log directory.

        :param log_dir: The tensorboard log path for output
        :type log_dir: str
        :param write_graph: If True, the model graph will be written using the TensorboardX library
        :type write_graph: bool
        :param write_batch_metrics: If True, batch metrics will be written
        :type write_batch_metrics: bool
        :param batch_step_size: The step size to use when writing batch metrics, make this larger to reduce latency
        :type batch_step_size: int
        :param write_epoch_metrics: If True, metrics from the end of the epoch will be written
        :type write_epoch_metrics: True
        :param comment: Descriptive comment to append to path
        :type comment: str
        """
        super(TensorBoard, self).__init__()

        self.log_dir = log_dir
        self.write_graph = write_graph
        self.write_batch_metrics = write_batch_metrics
        self.batch_step_size = batch_step_size
        self.write_epoch_metrics = write_epoch_metrics
        self.comment = comment

        if self.write_graph:
            def handle_graph(state):
                dummy = torch.rand(state[torchbearer.X].size(), requires_grad=False)
                model = copy.deepcopy(state[torchbearer.MODEL]).to('cpu')
                self._writer.add_graph(model, (dummy, ))
                self._handle_graph = lambda _: ...
            self._handle_graph = handle_graph
        else:
            self._handle_graph = lambda _: ...

        self._writer = None
        self._batch_writer = None

    def on_start(self, state):
        self.log_dir = os.path.join(self.log_dir, state[torchbearer.MODEL].__class__.__name__ + '_' + self.comment)
        self._writer = SummaryWriter(log_dir=self.log_dir)

    def on_start_epoch(self, state):
        if self.write_batch_metrics:
            log_dir = os.path.join(self.log_dir, 'epoch-' + str(state[torchbearer.EPOCH]))
            self._batch_writer = SummaryWriter(log_dir=log_dir)

    def on_end(self, state):
        self._writer.close()

    def on_sample(self, state):
        self._handle_graph(state)

    def on_step_training(self, state):
        if self.write_batch_metrics and state[torchbearer.BATCH] % self.batch_step_size == 0:
            for metric in state[torchbearer.METRICS]:
                self._batch_writer.add_scalar('batch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.BATCH])

    def on_step_validation(self, state):
        if self.write_batch_metrics and state[torchbearer.BATCH] % self.batch_step_size == 0:
            for metric in state[torchbearer.METRICS]:
                self._batch_writer.add_scalar('batch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.BATCH])

    def on_end_epoch(self, state):
        if self.write_batch_metrics:
            self._batch_writer.close()

        if self.write_epoch_metrics:
            for metric in state[torchbearer.METRICS]:
                self._writer.add_scalar('epoch/' + metric, state[torchbearer.METRICS][metric], state[torchbearer.EPOCH])


class TensorBoardImages(Callback):
    """The TensorBoardImages callback will write a selection of images from the validation pass to tensorboard using the
    TensorboardX library and torchvision.utils.make_grid
    """

    def __init__(self, log_dir='./logs',
                 comment='torchbearer',
                 name='Image',
                 key=torchbearer.Y_PRED,
                 write_each_epoch=True,
                 num_images=16,
                 nrow=8,
                 padding=2,
                 normalize=False,
                 range=None,
                 scale_each=False,
                 pad_value=0):
        """Create TensorBoardImages callback which writes images from the given key to the given path. Full name of
        image sub directory will be model name + _ + comment.

        :param log_dir: The tensorboard log path for output
        :type log_dir: str
        :param comment: Descriptive comment to append to path
        :type comment: str
        :param name: The name of the image
        :type name: str
        :param key: The key in state containing image data (tensor of size [c, w, h] or [b, c, w, h])
        :type key: str
        :param write_each_epoch: If True, write data on every epoch, else write only for the first epoch.
        :type write_each_epoch: bool
        :param num_images: The number of images to write
        :type num_images: int
        :param nrow: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        :param padding: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        :param normalize: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        :param range: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        :param scale_each: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        :param pad_value: See `torchvision.utils.make_grid https://pytorch.org/docs/stable/torchvision/utils.html#torchvision.utils.make_grid`
        """
        self.log_dir = log_dir
        self.comment = comment
        self.name = name
        self.key = key
        self.write_each_epoch = write_each_epoch
        self.num_images = num_images
        self.nrow = nrow
        self.padding = padding
        self.normalize = normalize
        self.range = range
        self.scale_each = scale_each
        self.pad_value = pad_value

        self._writer = None
        self._data = None
        self.done = False

    def on_start(self, state):
        log_dir = os.path.join(self.log_dir, state[torchbearer.MODEL].__class__.__name__ + '_' + self.comment)
        self._writer = SummaryWriter(log_dir=log_dir)

    def on_step_validation(self, state):
        if not self.done:
            data = state[self.key].clone()

            if len(data.size()) == 3:
                data = data.unsqueeze(1)

            if self._data is None:
                remaining = self.num_images if self.num_images < data.size(0) else data.size(0)

                self._data = data[:remaining].to('cpu')
            else:
                remaining = self.num_images - self._data.size(0)

                if remaining > data.size(0):
                    remaining = data.size(0)

                self._data = torch.cat((self._data, data[:remaining].to('cpu')), dim=0)

            if self._data.size(0) >= self.num_images:
                image = utils.make_grid(
                    self._data,
                    nrow=self.nrow,
                    padding=self.padding,
                    normalize=self.normalize,
                    range=self.range,
                    scale_each=self.scale_each,
                    pad_value=self.pad_value
                )
                self._writer.add_image(self.name, image, state[torchbearer.EPOCH])
                self.done = True
                self._data = None

    def on_end_epoch(self, state):
        if self.write_each_epoch:
            self.done = False

    def on_end(self, state):
        self._writer.close()


class TensorBoardProjector(Callback):
    """The TensorBoardProjector callback is used to write images from the validation pass to Tensorboard using the
    TensorboardX library.
    """

    def __init__(self, log_dir='./logs',
                 comment='torchbearer',
                 num_images=100,
                 avg_pool_size=1,
                 avg_data_channels=True,
                 write_data=True,
                 write_features=True,
                 features_key=torchbearer.Y_PRED):
        """Construct a TensorBoardProjector callback which writes images to the given directory and, if required,
        associated features.

        :param log_dir: The tensorboard log path for output
        :type log_dir: str
        :param comment: Descriptive comment to append to path
        :type comment: str
        :param num_images: The number of images to write
        :type num_images: int
        :param avg_pool_size: Size of the average pool to perform on the image. This is recommended to reduce the
        overall image sizes and improve latency
        :type avg_pool_size: int
        :param avg_data_channels: If True, the image data will be averaged in the channel dimension
        :type avg_data_channels: bool
        :param write_data: If True, the raw data will be written as an embedding
        :type write_data: bool
        :param write_features: If True, the image features will be written as an embedding
        :type write_features: bool
        :param features_key: The key in state to use for the embedding. Typically model output but can be used to show
        features from any layer of the model.
        :type features_key: str
        """
        self.log_dir = log_dir
        self.comment = comment
        self.num_images = num_images
        self.avg_pool_size = avg_pool_size
        self.avg_data_channels = avg_data_channels
        self.write_data = write_data
        self.write_features = write_features
        self.features_key = features_key

        self._writer = None

        self.done = False

    def on_start(self, state):
        log_dir = os.path.join(self.log_dir, state[torchbearer.MODEL].__class__.__name__ + '_' + self.comment)
        self._writer = SummaryWriter(log_dir=log_dir)

    def on_step_validation(self, state):
        if not self.done:
            x = state[torchbearer.X].data.clone()

            if len(x.size()) == 3:
                x = x.unsqueeze(1)

            x = F.avg_pool2d(x, self.avg_pool_size).data

            data = None

            if state[torchbearer.EPOCH] == 0 and self.write_data:
                if self.avg_data_channels:
                    data = torch.mean(x, 1)
                else:
                    data = x

                data = data.view(data.size(0), -1)

            feature = None

            if self.write_features:
                feature = state[self.features_key].data.clone()
                feature = feature.view(feature.size(0), -1)

            label = state[torchbearer.Y_TRUE].data.clone()

            if state[torchbearer.BATCH] == 0:
                remaining = self.num_images if self.num_images < label.size(0) else label.size(0)

                self._images = x[:remaining].to('cpu')
                self._labels = label[:remaining].to('cpu')

                if data is not None:
                    self._data = data[:remaining].to('cpu')

                if feature is not None:
                    self._features = feature[:remaining].to('cpu')
            else:
                remaining = self.num_images - self._labels.size(0)

                if remaining > label.size(0):
                    remaining = label.size(0)

                self._images = torch.cat((self._images, x[:remaining].to('cpu')), dim=0)
                self._labels = torch.cat((self._labels, label[:remaining].to('cpu')), dim=0)

                if data is not None:
                    self._data = torch.cat((self._data, data[:remaining].to('cpu')), dim=0)

                if feature is not None:
                    self._features = torch.cat((self._features, feature[:remaining].to('cpu')), dim=0)

            if self._labels.size(0) >= self.num_images:
                if state[torchbearer.EPOCH] == 0 and self.write_data:
                    self._writer.add_embedding(self._data, metadata=self._labels, label_img=self._images, tag='data', global_step=-1)
                if self.write_features:
                    self._writer.add_embedding(self._features, metadata=self._labels, label_img=self._images, tag='features', global_step=state[torchbearer.EPOCH])
                self.done = True

    def on_end_epoch(self, state):
        if self.write_features:
            self.done = False

    def on_end(self, state):
        self._writer.close()



import unittest

from unittest.mock import Mock, call

from torchbearer.metrics import RunningMean, Metric, RunningMetric, Mean, Std

import torch


class TestStd(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([0.1, 0.2, 0.3]),
                                            torch.FloatTensor([0.4, 0.5, 0.6]),
                                            torch.FloatTensor([0.7, 0.8, 0.9]),
                                            torch.ones(torch.Size([]))]

        self._std = Std('test')
        self._std.reset({})
        self._target = 0.31622776601684

    def test_train(self):
        self._std.train()
        for i in range(5):
            self._std.process(self._metric.process())
        result = self._std.process_final({})
        self.assertAlmostEqual(self._target, result)

    def test_validate(self):
        self._std.eval()
        for i in range(5):
            self._std.process(self._metric.process())
        result = self._std.process_final({})
        self.assertAlmostEqual(self._target, result)


class TestMean(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([0.1, 0.2, 0.3]),
                                            torch.FloatTensor([0.4, 0.5, 0.6]),
                                            torch.FloatTensor([0.7, 0.8, 0.9]),
                                            torch.ones(torch.Size([]))]

        self._mean = Mean('test')
        self._mean.reset({})
        self._target = 0.5

    def test_train_dict(self):
        self._mean.train()
        for i in range(5):
            self._mean.process(self._metric.process())
        result = self._mean.process_final({})
        self.assertAlmostEqual(self._target, result)

    def test_validate_dict(self):
        self._mean.eval()
        for i in range(5):
            self._mean.process(self._metric.process())
        result = self._mean.process_final({})
        self.assertAlmostEqual(self._target, result)


class TestRunningMetric(unittest.TestCase):
    def setUp(self):
        self._metric = RunningMetric('test', batch_size=5, step_size=5)
        self._metric.reset({})
        self._metric._process_train = Mock(return_value=3)
        self._metric._step = Mock(return_value='output')

    def test_train_called_with_state(self):
        self._metric.train()
        self._metric.process({'test': -1})
        self._metric._process_train.assert_called_with({'test': -1})

    def test_cache_one_step(self):
        self._metric.train()
        for i in range(6):
            self._metric.process({})
        self._metric._step.assert_has_calls([call([3]), call([3, 3, 3, 3, 3])])

    def test_empty_methods(self):
        metric = RunningMetric('test')
        self.assertTrue(metric._step(['test']) is None)
        self.assertTrue(metric._process_train(['test']) is None)


class TestRunningMean(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._mean = RunningMean('test')
        self._cache = [1.0, 1.5, 2.0]
        self._target = 1.5

    def test_train(self):
        result = self._mean._process_train(torch.FloatTensor([1.0, 1.5, 2.0]))
        self.assertAlmostEqual(self._target, result, 3, 0.002)

    def test_step(self):
        result = self._mean._step(self._cache)
        self.assertEqual(self._target, result)


"""
..  automodule:: torchbearer.torchbearer
        :members:
        :undoc-members:

..  automodule:: torchbearer.state
        :members:
        :undoc-members:

..  automodule:: torchbearer.cv_utils
        :members:
        :undoc-members:
"""

from .state import *
from . import metrics
from . import callbacks
from .torchbearer import *
from . import cv_utils



from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from torchbearer.callbacks import TerminateOnNaN


class TestTerminateOnNaN(TestCase):

    @patch('sys.stdout', new_callable=StringIO)
    def test_should_terminate(self, _):
        state = {
            'epoch': 0,
            't': 0,
            'stop_training': False,
            'metrics': {'running_loss': float('nan')}
        }

        terminator = TerminateOnNaN()

        terminator.on_step_training(state)
        self.assertTrue(state['stop_training'])
        terminator.on_step_validation(state)
        self.assertTrue(state['stop_training'])
        terminator.on_end_epoch(state)
        self.assertTrue(state['stop_training'])

    def test_should_not_terminate(self):
        state = {
            'epoch': 0,
            't': 0,
            'stop_training': False,
            'metrics': {'running_loss': 0.01}
        }

        terminator = TerminateOnNaN()

        terminator.on_step_training(state)
        self.assertFalse(state['stop_training'])
        terminator.on_step_validation(state)
        self.assertFalse(state['stop_training'])
        terminator.on_end_epoch(state)
        self.assertFalse(state['stop_training'])

    @patch('sys.stdout', new_callable=StringIO)
    def test_monitor_should_terminate(self, _):
        state = {
            'epoch': 0,
            't': 0,
            'stop_training': False,
            'metrics': {'running_loss': 0.001, 'test_metric': float('nan')}
        }

        terminator = TerminateOnNaN(monitor='test_metric')

        terminator.on_step_training(state)
        self.assertTrue(state['stop_training'])
        terminator.on_step_validation(state)
        self.assertTrue(state['stop_training'])
        terminator.on_end_epoch(state)
        self.assertTrue(state['stop_training'])

    def test_monitor_should_not_terminate(self):
        state = {
            'epoch': 0,
            't': 0,
            'stop_training': False,
            'metrics': {'running_loss': 0.01, 'test_metric': 0.001}
        }

        terminator = TerminateOnNaN(monitor='test_metric')

        terminator.on_step_training(state)
        self.assertFalse(state['stop_training'])
        terminator.on_step_validation(state)
        self.assertFalse(state['stop_training'])
        terminator.on_end_epoch(state)
        self.assertFalse(state['stop_training'])

    @patch('sys.stdout', new_callable=StringIO)
    def test_not_found_metric(self, _):
        state = {
            'epoch': 0,
            't': 0,
            'stop_training': False,
            'metrics': {'running_loss': 0.001, 'test_metric': float('nan')}
        }

        terminator = TerminateOnNaN(monitor='test')

        terminator.on_step_training(state)
        self.assertFalse(state['stop_training'])
        terminator.on_step_validation(state)
        self.assertFalse(state['stop_training'])
        terminator.on_end_epoch(state)
        self.assertFalse(state['stop_training'])



import torchbearer

import torch

from torchbearer.callbacks.callbacks import Callback
import os


class _Checkpointer(Callback):
    def __init__(self, fileformat, pickle_module=torch.serialization.pickle, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
        super().__init__()
        self.fileformat = fileformat

        self.pickle_module = pickle_module
        self.pickle_protocol = pickle_protocol

        self.most_recent = None

        if fileformat.__contains__(os.sep) and not os.path.exists(os.path.dirname(fileformat)):
            os.makedirs(os.path.dirname(fileformat))

    def save_checkpoint(self, model_state, overwrite_most_recent=False):
        state = {}
        state.update(model_state)
        state.update(model_state[torchbearer.METRICS])

        filepath = self.fileformat.format(**state)

        if self.most_recent is not None and overwrite_most_recent:
            os.remove(self.most_recent)

        torch.save(model_state[torchbearer.SELF].state_dict(), filepath, pickle_module=self.pickle_module,
                   pickle_protocol=self.pickle_protocol)

        self.most_recent = filepath


def ModelCheckpoint(filepath='model.{epoch:02d}-{val_loss:.2f}.pt',
        monitor='val_loss', save_best_only=False, mode='auto', period=1, min_delta=0):
    """Save the model after every epoch.
    `filepath` can contain named formatting options,
    which will be filled any values from state.
    For example: if `filepath` is `weights.{epoch:02d}-{val_loss:.2f}`,
    then the model checkpoints will be saved with the epoch number and
    the validation loss in the filename. The torch model will be saved to filename.pt
    and the torchbearermodel state will be saved to filename.torchbearer.

    :param filepath: Path to save the model file
    :type filepath: str
    :param monitor: Quantity to monitor
    :type monitor: str
    :param save_best_only: If `save_best_only=True`, the latest best model according to the quantity monitored will not\
    be overwritten
    :type save_best_only: bool
    :param mode: One of {auto, min, max}. If `save_best_only=True`, the decision to overwrite the current save file is\
    made based on either the maximization or the minimization of the monitored quantity. For `val_acc`, this should be\
    `max`, for `val_loss` this should be `min`, etc. In `auto` mode, the direction is automatically inferred from the\
    name of the monitored quantity.
    :type mode: str
    :param period: Interval (number of epochs) between checkpoints
    :type period: int
    :param min_delta: If `save_best_only=True`, this is the minimum improvement required to trigger a save
    :type min_delta: float
    """
    if save_best_only:
        check = Best(filepath, monitor, mode, period, min_delta)
    else:
        check = Interval(filepath, period)

    return check


class MostRecent(_Checkpointer):
    """Model checkpointer which saves the most recent model.
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', pickle_module=torch.serialization.pickle,
                 pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
        """Create a model checkpointer which saves the most recent model to the given filepath.

        :param filepath: Path to save the model file
        :type filepath: str
        :param pickle_module: The pickle module to use, default is 'torch.serialization.pickle'
        :param pickle_protocol: The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'
        """
        super().__init__(filepath, pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.filepath = filepath

    def on_end_epoch(self, model_state):
        super().on_end_training(model_state)
        self.save_checkpoint(model_state, overwrite_most_recent=True)


class Best(_Checkpointer):
    """Model checkpointer which saves the best model according to a metric.
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', monitor='val_loss', mode='auto', period=1,
                 min_delta=0, pickle_module=torch.serialization.pickle,
                 pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
        """Create a model checkpointer which saves the best model according to the given configurations.

        :param filepath: Path to save the model file
        :type filepath: str
        :param monitor: Quantity to monitor
        :type monitor: str
        :param mode: One of {auto, min, max}. The decision to overwrite the current save file is made based on either\
        the maximization or the minimization of the monitored quantity. For `val_acc`, this should be `max`, for\
        `val_loss` this should be `min`, etc. In `auto` mode, the direction is automatically inferred from the name of\
        the monitored quantity.
        :type mode: str
        :param period: Interval (number of epochs) between checkpoints
        :type period: int
        :param min_delta: This is the minimum improvement required to trigger a save
        :type min_delta: float
        :param pickle_module: The pickle module to use, default is 'torch.serialization.pickle'
        :param pickle_protocol: The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'
        """
        super().__init__(filepath, pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.min_delta = min_delta
        self.mode = mode
        self.monitor = monitor
        self.period = period
        self.epochs_since_last_save = 0

        if self.mode not in ['min', 'max']:
            if 'acc' in self.monitor:
                self.mode = 'max'
            else:
                self.mode = 'min'

        if self.mode == 'min':
            self.min_delta *= -1
            self.monitor_op = lambda x1, x2: (x1-self.min_delta) < x2
        elif self.mode == 'max':
            self.min_delta *= 1
            self.monitor_op = lambda x1, x2: (x1-self.min_delta) > x2

    def on_start(self, state):
        self.best = float('inf') if self.mode == 'min' else -float('inf')

    def on_end_epoch(self, model_state):

        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0

            current = model_state[torchbearer.METRICS][self.monitor]

            if self.monitor_op(current, self.best):
                self.best = current
                self.save_checkpoint(model_state, overwrite_most_recent=True)


class Interval(_Checkpointer):
    """Model checkpointer which saves the model every given number of epochs.
    """

    def __init__(self, filepath='model.{epoch:02d}-{val_loss:.2f}.pt', period=1, pickle_module=torch.serialization.pickle, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
        """Create a model checkpointer which saves the model every 'period' epochs to the given filepath.

        :param filepath: Path to save the model file
        :type filepath: str
        :param period: Interval (number of epochs) between checkpoints
        :type period: int
        :param pickle_module: The pickle module to use, default is 'torch.serialization.pickle'
        :param pickle_protocol: The pickle protocol to use, default is 'torch.serialization.DEFAULT_PROTOCOL'
        """
        super().__init__(filepath, pickle_module=pickle_module, pickle_protocol=pickle_protocol)
        self.period = period
        self.epochs_since_last_save = 0

    def on_end_epoch(self, model_state):
        super().on_end_training(model_state)

        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            self.save_checkpoint(model_state)






from setuptools import setup

setup(
    name='torchbearer',
    version='0.1.5',
    packages=['torchbearer', 'torchbearer.metrics', 'torchbearer.callbacks', 'tests', 'tests.metrics', 'tests.callbacks'],
    url='https://github.com/ecs-vlc/torchbearer',
    download_url='https://github.com/ecs-vlc/torchbearer/archive/0.1.5.tar.gz',
    license='GPL-3.0',
    author='Matt Painter',
    author_email='mp2u16@ecs.soton.ac.uk',
    description='A model training library for pytorch',
    install_requires=['numpy', 'torch>=0.4', 'torchvision', 'scipy', 'scikit-learn', 'tqdm', 'tensorboardX>=1.2'],
    python_requires='>=3',
)



import unittest
from unittest.mock import Mock

from torchbearer.metrics import MetricFactory, MetricList, Metric, MetricTree, AdvancedMetric


class TestMetricFactory(unittest.TestCase):
    def test_empty_build(self):
        factory = MetricFactory()
        self.assertTrue(factory.build() is None)


class TestMetricTree(unittest.TestCase):
    def test_process(self):
        root = Metric('test')
        root.process = Mock(return_value='test')
        leaf1 = Metric('test')
        leaf1.process = Mock(return_value={'test': 10})
        leaf2 = Metric('test')
        leaf2.process = Mock(return_value=None)

        tree = MetricTree(root)
        tree.add_child(leaf1)
        tree.add_child(leaf2)

        self.assertTrue(tree.process('args') == {'test': 10})

        root.process.assert_called_once_with('args')
        leaf1.process.assert_called_once_with('test')
        leaf2.process.assert_called_once_with('test')

    def test_process_final(self):
        root = Metric('test')
        root.process_final = Mock(return_value='test')
        leaf1 = Metric('test')
        leaf1.process_final = Mock(return_value={'test': 10})
        leaf2 = Metric('test')
        leaf2.process_final = Mock(return_value=None)

        tree = MetricTree(root)
        tree.add_child(leaf1)
        tree.add_child(leaf2)

        self.assertTrue(tree.process_final('args') == {'test': 10})

        root.process_final.assert_called_once_with('args')
        leaf1.process_final.assert_called_once_with('test')
        leaf2.process_final.assert_called_once_with('test')

    def test_train(self):
        root = Metric('test')
        root.train = Mock()
        leaf = Metric('test')
        leaf.train = Mock()

        tree = MetricTree(root)
        tree.add_child(leaf)

        tree.train()
        root.train.assert_called_once()
        leaf.train.assert_called_once()

    def test_eval(self):
        root = Metric('test')
        root.eval = Mock()
        leaf = Metric('test')
        leaf.eval = Mock()

        tree = MetricTree(root)
        tree.add_child(leaf)

        tree.eval()
        root.eval.assert_called_once()
        leaf.eval.assert_called_once()

    def test_reset(self):
        root = Metric('test')
        root.reset = Mock()
        leaf = Metric('test')
        leaf.reset = Mock()

        tree = MetricTree(root)
        tree.add_child(leaf)

        tree.reset({})
        root.reset.assert_called_once_with({})
        leaf.reset.assert_called_once_with({})


class TestMetricList(unittest.TestCase):
    def test_list_in_list(self):
        metric = MetricList(['acc', MetricList(['loss'])])
        self.assertTrue(metric.metric_list[0].name == 'acc')
        self.assertTrue(metric.metric_list[1].name == 'loss')

    def test_default_acc(self):
        metric = MetricList(['acc'])
        self.assertTrue(metric.metric_list[0].name == 'acc', msg='acc not in: ' + str(metric.metric_list))

    def test_default_loss(self):
        metric = MetricList(['loss'])
        self.assertTrue(metric.metric_list[0].name == 'loss', msg='loss not in: ' + str(metric.metric_list))

    def test_default_epoch(self):
        metric = MetricList(['epoch'])
        self.assertTrue(metric.metric_list[0].name == 'epoch', msg='loss not in: ' + str(metric.metric_list))

    def test_process(self):
        my_mock = Metric('test')
        my_mock.process = Mock(return_value={'test': -1})
        metric = MetricList([my_mock])
        result = metric.process({'state': -1})
        self.assertEqual({'test': -1}, result)
        my_mock.process.assert_called_once_with({'state': -1})

    def test_process_final(self):
        my_mock = Metric('test')
        my_mock.process_final = Mock(return_value={'test': -1})
        metric = MetricList([my_mock])
        result = metric.process_final({'state': -1})
        self.assertEqual({'test': -1}, result)
        my_mock.process_final.assert_called_once_with({'state': -1})

    def test_train(self):
        my_mock = Metric('test')
        my_mock.train = Mock(return_value=None)
        metric = MetricList([my_mock])
        metric.train()
        my_mock.train.assert_called_once()

    def test_eval(self):
        my_mock = Metric('test')
        my_mock.eval = Mock(return_value=None)
        metric = MetricList([my_mock])
        metric.eval()
        my_mock.eval.assert_called_once()

    def test_reset(self):
        my_mock = Metric('test')
        my_mock.reset = Mock(return_value=None)
        metric = MetricList([my_mock])
        metric.reset({'state': -1})
        my_mock.reset.assert_called_once_with({'state': -1})


class TestAdvancedMetric(unittest.TestCase):
    def test_empty_methods(self):
        metric = AdvancedMetric('test')

        self.assertTrue(metric.process_train() is None)
        self.assertTrue(metric.process_final_train() is None)
        self.assertTrue(metric.process_validate() is None)
        self.assertTrue(metric.process_final_validate() is None)

    def test_train(self):
        metric = AdvancedMetric('test')
        metric.process_train = Mock()
        metric.process_final_train = Mock()

        metric.train()
        metric.process('testing')
        metric.process_train.assert_called_once_with('testing')

        metric.process_final('testing')
        metric.process_final_train.assert_called_once_with('testing')

    def test_eval(self):
        metric = AdvancedMetric('test')
        metric.process_validate = Mock()
        metric.process_final_validate = Mock()

        metric.eval()
        metric.process('testing')
        metric.process_validate.assert_called_once_with('testing')

        metric.process_final('testing')
        metric.process_final_validate.assert_called_once_with('testing')



import warnings

import torch
from torch.utils.data import DataLoader, TensorDataset

import torchbearer
from torchbearer import metrics as torchbearer_metrics
from torchbearer.callbacks.aggregate_predictions import AggregatePredictions
from torchbearer.callbacks.callbacks import CallbackList
from torchbearer.callbacks.printer import Tqdm


class Model:
    """ Torchbearermodel to wrap base torch model and provide training environment around it
    """
    def __init__(self, model, optimizer, loss_criterion, metrics=[]):
        """ Create torchbearermodel which wraps a base torchmodel and provides a training environment surrounding it

        :param model: The base pytorch model
        :type model: torch.nn.Module
        :param optimizer: The optimizer used for pytorch model weight updates
        :type optimizer: torch.optim.Optimizer
        :param loss_criterion: The final loss criterion that provides a loss value to the optimizer
        :type loss_criterion: function
        :param metrics: Additional metrics for display and use within callbacks
        :type metrics: list
        """
        super().__init__()
        self.main_state = {
            torchbearer.MODEL: model,
            torchbearer.CRITERION: loss_criterion,
            torchbearer.OPTIMIZER: optimizer,
            torchbearer.DEVICE: 'cpu',
            torchbearer.DATA_TYPE: torch.float32,
            torchbearer.METRIC_LIST: torchbearer_metrics.MetricList(metrics),
            torchbearer.SELF: self
        }

    def fit(self, x, y, batch_size=None, epochs=1, verbose=1, callbacks=[], validation_split=None,
            validation_data=None, shuffle=True, initial_epoch=0,
            steps_per_epoch=None, validation_steps=None, workers=1, pass_state=False):
        """ Perform fitting of a model to given data and label tensors

        :param x: The input data tensor
        :type x: torch.Tensor
        :param y: The target labels for data tensor x
        :type y: torch.Tensor
        :param batch_size: The mini-batch size (number of samples processed for a single weight update)
        :type batch_size: int
        :param epochs: The number of training epochs to be run (each sample from the dataset is viewed exactly once)
        :type epochs: int
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param callbacks: The list of torchbearer callbacks to be called during training and validation
        :type callbacks: list
        :param validation_split: Fraction of the training dataset to be set aside for validation testing
        :type validation_split: float
        :param validation_data: Optional validation data tensor
        :type validation_data: (torch.Tensor, torch.Tensor)
        :param shuffle: If True mini-batches of training/validation data are randomly selected, if False mini-batches samples are selected in order defined by dataset
        :type shuffle: bool
        :param initial_epoch: The integer value representing the first epoch - useful for continuing training after a number of epochs
        :type initial_epoch: int
        :param steps_per_epoch: The number of training mini-batches to run per epoch
        :type steps_per_epoch: int
        :param validation_steps: The number of validation mini-batches to run per epoch
        :type validation_steps: int
        :param workers: The number of cpu workers devoted to batch loading and aggregating
        :type workers: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: The final state context dictionary
        :rtype: dict[str,any]
        """
        
        trainset, valset = torchbearer.cv_utils.get_train_valid_sets(x, y, validation_data, validation_split, shuffle=shuffle)
        trainloader = DataLoader(trainset, batch_size, shuffle=shuffle, num_workers=workers)

        if valset is not None:
            valloader = DataLoader(valset, batch_size, shuffle=shuffle, num_workers=workers)
        else:
            valloader = None

        return self.fit_generator(trainloader, train_steps=steps_per_epoch, epochs=epochs, verbose=verbose,
                                  callbacks=callbacks, validation_generator=valloader, validation_steps=validation_steps,
                                  initial_epoch=initial_epoch, pass_state=pass_state)

    def fit_generator(self, generator, train_steps=None, epochs=1, verbose=1, callbacks=[],
                      validation_generator=None, validation_steps=None, initial_epoch=0, pass_state=False):
        """ Perform fitting of a model to given data generator

        :param generator: The training data generator (usually a pytorch DataLoader)
        :type generator: DataLoader
        :param train_steps: The number of training mini-batches to run per epoch
        :type train_steps: int
        :param epochs: The number of training epochs to be run (each sample from the dataset is viewed exactly once)
        :type epochs: int
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param callbacks: The list of torchbearer callbacks to be called during training and validation
        :type callbacks: list
        :param validation_generator: The validation data generator (usually a pytorch DataLoader)
        :type validation_generator: DataLoader
        :param validation_steps: The number of validation mini-batches to run per epoch
        :type validation_steps: int
        :param initial_epoch: The integer value representing the first epoch - useful for continuing training after a number of epochs
        :type initial_epoch: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: The final state context dictionary
        :rtype: dict[str,any]
        """
        if verbose == 1:
            callbacks = [Tqdm()] + callbacks
        _callbacks = CallbackList(callbacks)

        # Get train and validation steps
        if validation_steps is None and validation_generator is not None:
            validation_steps = len(validation_generator)
        if train_steps is None:
            train_steps = len(generator)
        if generator is not None and train_steps > len(generator):
            train_steps = len(generator)
        if not isinstance(train_steps, int):
            train_steps = int(train_steps)
            warnings.warn("Number of training steps is not an int, converting to int")

        if not isinstance(epochs, int):
            if isinstance(epochs, float):
                epochs = int(epochs)
                warnings.warn("Number of epochs is a float, converting to int")
            else:
                warnings.warn("Number of epochs is neither float nor int, setting to 0")
                epochs = 0

        # Init state
        state = {
            torchbearer.MAX_EPOCHS: epochs,
            torchbearer.TRAIN_STEPS: train_steps,
            torchbearer.BATCH: 0,
            torchbearer.GENERATOR: generator,
            torchbearer.STOP_TRAINING: False
        }
        state.update(self.main_state)

        _callbacks.on_start(state)

        for state[torchbearer.EPOCH] in range(initial_epoch, epochs):
            _callbacks.on_start_epoch(state)

            if state[torchbearer.GENERATOR] is not None:
                state[torchbearer.TRAIN_ITERATOR] = iter(state[torchbearer.GENERATOR])
            self.train()

            _callbacks.on_start_training(state)
            state[torchbearer.METRIC_LIST].reset(state)
            state[torchbearer.METRICS] = {}

            for state[torchbearer.BATCH] in range(0, state[torchbearer.TRAIN_STEPS]):
                # Extract batch
                if state[torchbearer.GENERATOR] is None: # TODO: Replace with flag check
                    self._load_batch_none('train', state)
                else:
                    self._load_batch_standard('train', state)

                _callbacks.on_sample(state)

                # Zero grads
                state[torchbearer.OPTIMIZER].zero_grad()

                # Forward pass
                if pass_state:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X], state=state)
                else:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X])
                _callbacks.on_forward(state)

                # Loss Calculation
                state[torchbearer.LOSS] = state[torchbearer.CRITERION](state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE])

                _callbacks.on_criterion(state)
                state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state)

                # Backwards pass
                state[torchbearer.LOSS].backward()
                _callbacks.on_backward(state)

                # Update parameters
                state[torchbearer.OPTIMIZER].step()
                _callbacks.on_step_training(state)

                if state[torchbearer.STOP_TRAINING]:
                    break

            state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state))
            final_metrics = state[torchbearer.METRICS]

            _callbacks.on_end_training(state)

            # Validate
            if validation_generator is not None or validation_steps is not None:
                state[torchbearer.VALIDATION_GENERATOR] = validation_generator
                state[torchbearer.VALIDATION_STEPS] = validation_steps
                self.eval()
                self._validate(state, _callbacks, pass_state)

            final_metrics.update(state[torchbearer.METRICS])
            state[torchbearer.METRICS] = final_metrics
            _callbacks.on_end_epoch(state)

            if state[torchbearer.STOP_TRAINING]:
                break
        _callbacks.on_end(state)

        return state

    def _test_loop(self, state, callbacks, pass_state, batch_loader, num_steps=None):
        """ The generic testing loop used for validation, evaluation and prediction

        :param state: The current state context dictionary
        :type state: dict[str,any]
        :param callbacks: The list of torchbearer callbacks to be called during training and validation
        :type callbacks: CallbackList
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :param batch_loader: The batch loader to use
        :type batch_loader: function
        :param num_steps: The number of testing mini-batches to run
        :return: The state context dictionary
        :rtype: dict[str,any]
        """
        with torch.no_grad():
            state[torchbearer.METRIC_LIST].reset(state)
            state[torchbearer.METRICS] = {}

            if num_steps is None:
                num_steps = len(state[torchbearer.VALIDATION_GENERATOR])
            if state[torchbearer.VALIDATION_GENERATOR] is not None and num_steps > len(state[torchbearer.VALIDATION_GENERATOR]):
                num_steps = len(state[torchbearer.VALIDATION_GENERATOR])
            if not isinstance(num_steps, int):
                num_steps = int(num_steps)
                warnings.warn('Num test steps is not an int, converting to int.', Warning)

            state[torchbearer.VALIDATION_STEPS] = num_steps
            if state[torchbearer.VALIDATION_GENERATOR] is not None:
                state[torchbearer.VALIDATION_ITERATOR] = iter(state[torchbearer.VALIDATION_GENERATOR])

            callbacks.on_start_validation(state)

            for state[torchbearer.BATCH] in range(state[torchbearer.VALIDATION_STEPS]):
                # Load batch
                batch_loader('validation', state)
                callbacks.on_sample_validation(state)

                # Forward pass
                if pass_state:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X], state=state)
                else:
                    state[torchbearer.Y_PRED] = state[torchbearer.MODEL](state[torchbearer.X])
                callbacks.on_forward_validation(state)

                # Loss and metrics
                if torchbearer.Y_TRUE in state:
                    state[torchbearer.LOSS] = state[torchbearer.CRITERION](state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE])
                    callbacks.on_criterion_validation(state)
                    state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state)

                callbacks.on_step_validation(state)
                if state[torchbearer.STOP_TRAINING]:
                    break

            if torchbearer.Y_TRUE in state:
                state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state))
            callbacks.on_end_validation(state)

        return state

    def _validate(self, state, _callbacks, pass_state):
        """ Perform a validation loop

        :param state: The current context state dictionary
        :param _callbacks: The list of torchbearer callbacks to be called during validation loop
        :type callbacks: CallbackList
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: None
        :rtype: None
        """
        self._test_loop(state, _callbacks, pass_state, self._load_batch_standard, state[torchbearer.VALIDATION_STEPS])

    def evaluate(self, x=None, y=None, batch_size=32, verbose=1, steps=None, pass_state=False):
        """ Perform an evaluation loop on given data and label tensors to evaluate metrics

        :param x: The input data tensor
        :type x: torch.Tensor
        :param y: The target labels for data tensor x
        :type y: torch.Tensor
        :param batch_size: The mini-batch size (number of samples processed for a single weight update)
        :type batch_size: int
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param steps: The number of evaluation mini-batches to run
        :type steps: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: The dictionary containing final metrics
        :rtype: dict[str,any]
        """
        trainset = DataLoader(TensorDataset(x, y), batch_size, steps)
        return self.evaluate_generator(trainset, verbose, pass_state=pass_state)

    def evaluate_generator(self, generator, verbose=1, steps=None, pass_state=False):
        """ Perform an evaluation loop on given data generator to evaluate metrics

        :param generator: The evaluation data generator (usually a pytorch DataLoader)
        :type generator: DataLoader
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param steps: The number of evaluation mini-batches to run
        :type steps: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: The dictionary containing final metrics
        :rtype: dict[str,any]
        """

        state = {torchbearer.EPOCH: 0, torchbearer.MAX_EPOCHS: 1, torchbearer.STOP_TRAINING: False, torchbearer.VALIDATION_GENERATOR: generator}
        state.update(self.main_state)

        _callbacks = []
        if verbose == 1:
            _callbacks.append(Tqdm('e'))

        if state[torchbearer.VALIDATION_GENERATOR] is None:
            batch_loader = self._load_batch_none
        else:
            batch_loader = self._load_batch_standard

        self._test_loop(state, CallbackList(_callbacks), pass_state, batch_loader, steps)

        return state[torchbearer.METRICS]

    def predict(self, x=None, batch_size=32, verbose=1, steps=None, pass_state=False):
        """ Perform a prediction loop on given data tensor to predict labels

        :param x: The input data tensor
        :type x: torch.Tensor
        :param batch_size: The mini-batch size (number of samples processed for a single weight update)
        :type batch_size: int
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param steps: The number of evaluation mini-batches to run
        :type steps: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: Tensor of final predicted labels
        :rtype: torch.Tensor
        """
        pred_set = DataLoader(TensorDataset(x), batch_size, steps)
        return self.predict_generator(pred_set, verbose, pass_state=pass_state)

    def predict_generator(self, generator, verbose=1, steps=None, pass_state=False):
        """Perform a prediction loop on given data generator to predict labels

        :param generator: The prediction data generator (usually a pytorch DataLoader)
        :type generator: DataLoader
        :param verbose: If 1 use tqdm progress frontend, else display no training progress
        :type verbose: int
        :param steps: The number of evaluation mini-batches to run
        :type steps: int
        :param pass_state: If True the state dictionary is passed to the torch model forward method, if False only the input data is passed
        :type pass_state: bool
        :return: Tensor of final predicted labels
        :rtype: torch.Tensor
        """
        state = {torchbearer.EPOCH: 0, torchbearer.MAX_EPOCHS: 1, torchbearer.STOP_TRAINING: False, torchbearer.VALIDATION_GENERATOR: generator}
        state.update(self.main_state)

        _callbacks = [AggregatePredictions()]
        if verbose == 1:
            _callbacks.append(Tqdm('p'))
        self._test_loop(state, CallbackList(_callbacks), pass_state, self._load_batch_predict, steps)

        return state[torchbearer.FINAL_PREDICTIONS]

    def train(self):
        """ Set model and metrics to training mode
        """
        self.main_state[torchbearer.MODEL].train()
        self.main_state[torchbearer.METRIC_LIST].train()

    def eval(self):
        """ Set model and metrics to evaluation mode
        """
        self.main_state[torchbearer.MODEL].eval()
        self.main_state[torchbearer.METRIC_LIST].eval()

    def to(self, *args, **kwargs):
        """ Moves and/or casts the parameters and buffers.

        :param args: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_
        :param kwargs: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_
        :return: Self torchbearermodel
        :rtype: Model
        """
        self.main_state[torchbearer.MODEL].to(*args, **kwargs)

        for state in self.main_state[torchbearer.OPTIMIZER].state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(*args, **kwargs)

        self.main_state = Model._update_device_and_dtype_from_args(self.main_state, *args, **kwargs)

        return self

    def cuda(self, device=None):
        """ Moves all model parameters and buffers to the GPU.

        :param device: if specified, all parameters will be copied to that device
        :type device: int, optional
        :return: Self torchbearermodel
        :rtype: Model
        """
        if device is None:
            device = torch.cuda.current_device()
        return self.to('cuda:' + str(device))

    def cpu(self):
        """ Moves all model parameters and buffers to the CPU.

        :return: Self torchbearermodel
        :rtype: Model
        """
        return self.to('cpu')

    def load_state_dict(self, state_dict, **kwargs):
        """ Copies parameters and buffers from :func:`state_dict` into this module and its descendants.

        :param state_dict: A dict containing parameters and persistent buffers.
        :type state_dict: dict
        :param kwargs: See: `torch.nn.Module.load_state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.load_state_dict>`_
        """
        self.main_state[torchbearer.MODEL].load_state_dict(state_dict[torchbearer.MODEL], **kwargs)
        self.main_state[torchbearer.OPTIMIZER].load_state_dict(state_dict[torchbearer.OPTIMIZER])

    def state_dict(self, **kwargs):
        """
        :param kwargs: See: `torch.nn.Module.state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.state_dict>`_

        :return: A dict containing parameters and persistent buffers.
        :rtype: dict
        """
        state_dict = {
            torchbearer.MODEL: self.main_state[torchbearer.MODEL].state_dict(**kwargs),
            torchbearer.OPTIMIZER: self.main_state[torchbearer.OPTIMIZER].state_dict()
        }
        return state_dict

    @staticmethod
    def _deep_to(batch, device, dtype):
        """ Static method to call :func:`to` on tensors or tuples. All items in tuple will have :func:_deep_to called

        :param batch: The mini-batch which requires a :func:`to` call
        :type batch: tuple, list, torch.Tensor
        :param device: The desired device of the batch
        :type device: torch.device
        :param dtype: The desired datatype of the batch
        :type dtype: torch.dtype
        :return: The moved or casted batch
        :rtype: tuple, list, torch.Tensor
        """
        is_tuple = isinstance(batch, tuple)

        if isinstance(batch, list) or isinstance(batch, tuple):
            batch = list(batch)
            for i in range(len(batch)):
                batch[i] = Model._deep_to(batch[i], device, dtype)
            batch = tuple(batch) if is_tuple else batch
        else:
            if batch.dtype.is_floating_point:
                batch = batch.to(device, dtype)
            else:
                batch = batch.to(device)

        return batch

    @staticmethod
    def _load_batch_standard(iterator, state):
        """ Static method to load a standard (input data, target) tuple mini-batch from iterator into state

        :param iterator: Training or validation data iterator
        :type iterator: iterable
        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]
        """
        state[torchbearer.X], state[torchbearer.Y_TRUE] = Model._deep_to(next(state[iterator + '_iterator']), state[torchbearer.DEVICE], state[torchbearer.DATA_TYPE])

    @staticmethod
    def _load_batch_none(_, state):
        """Static method to load a none (none, none) tuple mini-batch into state

        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]
        """
        state[torchbearer.X], state[torchbearer.Y_TRUE] = None, None

    @staticmethod
    def _load_batch_predict(iterator, state):
        """ Static method to load a prediction (input data, target) or (input data) mini-batch from iterator into state

        :param iterator: Training or validation data iterator
        :type iterator: iterable
        :param state: The current state dict of the :class:`Model`.
        :type state: dict[str,any]
        """
        data = Model._deep_to(next(state[iterator + '_iterator']), state[torchbearer.DEVICE], state[torchbearer.DATA_TYPE])
        if isinstance(data, list) or isinstance(data, tuple):
            state[torchbearer.X], state[torchbearer.Y_TRUE] = data
        else:
            state[torchbearer.X] = data

    @staticmethod
    def _update_device_and_dtype_from_args(main_state, *args, **kwargs):
        """ static method to update a main state dictionary with new data type and device values
        
        :param main_state: The main state to update
        :type main_state: dict[str,any]
        :param args: Arguments to the :func:`Model.to` function
        :param kwargs: Keyword arguments to the :func:`Model.to` function
        :return: Updated main state dictionary
        :rtype: dict[str,any]
        """
        for key, val in kwargs.items():
            if key == torchbearer.DATA_TYPE:
                main_state[torchbearer.DATA_TYPE] = kwargs[torchbearer.DATA_TYPE]
            elif torchbearer.DEVICE in kwargs:
                main_state[torchbearer.DEVICE] = kwargs[torchbearer.DEVICE]

        for arg in args:
            if isinstance(arg, torch.dtype):
                main_state[torchbearer.DATA_TYPE] = arg
            else:
                main_state[torchbearer.DEVICE] = arg

        return main_state


import unittest

from unittest.mock import Mock, patch

import torchbearer
from torchbearer.metrics import RocAucScore, MetricList

import torch

import numpy as np


class TestRocAucScore(unittest.TestCase):
    @patch('sklearn.metrics')
    def test_one_hot(self, mock_sklearn_metrics):
        mock_sklearn_metrics.roc_auc_score = Mock()
        metric = RocAucScore(one_hot_classes=3, one_hot_offset=1).build()
        metric.reset({torchbearer.DEVICE: 'cpu', torchbearer.DATA_TYPE: torch.float32})
        res = metric.process({torchbearer.BATCH: 0, torchbearer.DEVICE: 'cpu', torchbearer.DATA_TYPE: torch.float32,
                        torchbearer.Y_TRUE: torch.LongTensor([1, 2, 3]),
                        torchbearer.Y_PRED: torch.FloatTensor([[0.0, 0.0, 0.0], [1.1, 1.1, 1.1], [2.2, 2.2, 2.2]])})
        self.assertTrue('roc_auc_score' in res)
        mock_sklearn_metrics.roc_auc_score.assert_called_once()
        self.assertTrue(np.array_equal(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
                                       mock_sklearn_metrics.roc_auc_score.call_args_list[0][0][0]))
        try:
            np.testing.assert_array_almost_equal(np.array([[0.0, 0.0, 0.0], [1.1, 1.1, 1.1], [2.2, 2.2, 2.2]]),
                                                                                   mock_sklearn_metrics.roc_auc_score.call_args_list[0][0][1])
        except AssertionError:
            self.fail('y_pred not correctly passed to sklearn')

    @patch('sklearn.metrics')
    def test_non_one_hot(self, mock_sklearn_metrics):
        mock_sklearn_metrics.roc_auc_score = Mock()
        metric = RocAucScore(one_hot_labels=False).build()
        metric.reset({torchbearer.DEVICE: 'cpu', torchbearer.DATA_TYPE: torch.float32})
        res = metric.process({torchbearer.BATCH: 0, torchbearer.DEVICE: 'cpu', torchbearer.DATA_TYPE: torch.float32,
                        torchbearer.Y_TRUE: torch.LongTensor([[1, 1, 1], [2, 2, 2], [3, 3, 3]]),
                        torchbearer.Y_PRED: torch.FloatTensor([[0.0, 0.0, 0.0], [1.1, 1.1, 1.1], [2.2, 2.2, 2.2]])})
        self.assertTrue('roc_auc_score' in res)
        mock_sklearn_metrics.roc_auc_score.assert_called_once()
        self.assertTrue(np.array_equal(np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]]),
                                       mock_sklearn_metrics.roc_auc_score.call_args_list[0][0][0]))
        try:
            np.testing.assert_array_almost_equal(np.array([[0.0, 0.0, 0.0], [1.1, 1.1, 1.1], [2.2, 2.2, 2.2]]),
                                                                                   mock_sklearn_metrics.roc_auc_score.call_args_list[0][0][1])
        except AssertionError:
            self.fail('y_pred not correctly passed to sklearn')

    def test_default_roc(self):
        mlist = MetricList(['roc_auc'])
        self.assertTrue(mlist.metric_list[0].name == 'roc_auc_score')

        mlist = MetricList(['roc_auc_score'])
        self.assertTrue(mlist.metric_list[0].name == 'roc_auc_score')



import math

import torch
from torch.utils.data import TensorDataset, Dataset
import random


def train_valid_splitter(x, y, split, shuffle=True):
    """ Generate training and validation tensors from whole dataset data and label tensors
    
    :param x: Data tensor for whole dataset
    :type x: torch.Tensor
    :param y: Label tensor for whole dataset
    :type y: torch.Tensor
    :param split: Fraction of dataset to be used for validation
    :type split: float
    :param shuffle: If True randomize tensor order before splitting else do not randomize 
    :type shuffle: bool
    :return: Training and validation tensors (training data, training labels, validation data, validation labels)
    :rtype: tuple
    """
    num_samples_x = x.size()[0]
    num_valid_samples = math.floor(num_samples_x * split)

    if shuffle:
        indicies = torch.randperm(num_samples_x)
        x, y = x[indicies], y[indicies]

    x_val, y_val = x[:num_valid_samples], y[:num_valid_samples]
    x, y = x[num_valid_samples:], y[num_valid_samples:]

    return x, y, x_val, y_val


def get_train_valid_sets(x, y, validation_data, validation_split, shuffle=True):
    """ Generate validation and training datasets from whole dataset tensors
    
    :param x: Data tensor for dataset
    :type x: torch.Tensor
    :param y: Label tensor for dataset
    :type y: torch.Tensor
    :param validation_data: Optional validation data (x_val, y_val) to be used instead of splitting x and y tensors 
    :type validation_data: (torch.Tensor, torch.Tensor)
    :param validation_split: Fraction of dataset to be used for validation
    :type validation_split: float
    :param shuffle: If True randomize tensor order before splitting else do not randomize 
    :type shuffle: bool
    :return: Training and validation datasets
    :rtype: tuple
    """

    valset = None

    if validation_data is not None:
        x_val, y_val = validation_data
    elif isinstance(validation_split, float):
        x, y, x_val, y_val = train_valid_splitter(x, y, validation_split, shuffle=shuffle)
    else:
        x_val, y_val = None, None

    trainset = TensorDataset(x, y)
    if x_val is not None and y_val is not None:
        valset = TensorDataset(x_val, y_val)

    return trainset, valset


class DatasetValidationSplitter:
    def __init__(self, dataset_len, split_fraction, shuffle_seed=None):
        """ Generates training and validation split indicies for a given dataset length and creates training and
        validation datasets using these splits

        :param dataset_len: The length of the dataset to be split into training and validation
        :param split_fraction: The fraction of the whole dataset to be used for validation
        :param shuffle_seed: Optional random seed for the shuffling process
        """
        super().__init__()
        self.dataset_len = dataset_len
        self.split_fraction = split_fraction
        self.valid_ids = None
        self.train_ids = None
        self._gen_split_ids(shuffle_seed)

    def _gen_split_ids(self, seed):
        all_ids = list(range(self.dataset_len))

        if seed is not None:
            random.seed(seed)
        random.shuffle(all_ids)

        num_valid_ids = math.floor(self.dataset_len*self.split_fraction)
        self.valid_ids = all_ids[:num_valid_ids]
        self.train_ids = all_ids[num_valid_ids:]

    def get_train_dataset(self, dataset):
        """ Creates a training dataset from existing dataset

        :param dataset: Dataset to be split into a training dataset
        :type dataset: torch.utils.data.Dataset
        :return: Training dataset split from whole dataset
        :rtype: torch.utils.data.Dataset
        """
        return SubsetDataset(dataset, self.train_ids)

    def get_val_dataset(self, dataset):
        """ Creates a validation dataset from existing dataset

        :param dataset: Dataset to be split into a validation dataset
        :type dataset: torch.utils.data.Dataset
        :return: Validation dataset split from whole dataset
        :rtype: torch.utils.data.Dataset
        """
        return SubsetDataset(dataset, self.valid_ids)


class SubsetDataset(Dataset):
    def __init__(self, dataset, ids):
        """ Dataset that consists of a subset of a previous dataset

        :param dataset: Complete dataset
        :type dataset: torch.utils.data.Dataset
        :param ids: List of subset IDs
        :type ids: list
        """
        super().__init__()
        self.dataset = dataset
        self.ids = ids

    def __getitem__(self, index):
        return self.dataset.__getitem__(self.ids[index])

    def __len__(self):
        return len(self.ids)




from unittest import TestCase
import torch
from torchbearer.callbacks import AggregatePredictions
import torchbearer as tb


class TestAggregatePredictions(TestCase):

    def test_aggreate_predictions(self):
        aggregator = AggregatePredictions()

        y_pred_1 = torch.Tensor([1,2,3])
        y_pred_2 = torch.Tensor([3,4,5])

        state_1 = {tb.Y_PRED: y_pred_1}
        state_2 = {tb.Y_PRED: y_pred_2}
        final_state = {}

        aggregator.on_step_validation(state_1)
        self.assertTrue(list(aggregator.predictions_list[0].numpy()) == list(y_pred_1.numpy()))

        aggregator.on_step_validation(state_2)
        self.assertTrue(list(aggregator.predictions_list[1].numpy()) == list(y_pred_2.numpy()))

        aggregate = torch.cat([y_pred_1, y_pred_2])
        aggregator.on_end_validation(final_state)
        self.assertTrue(list(final_state[tb.FINAL_PREDICTIONS].numpy()) == list(aggregate.numpy()))





"""
..  automodule:: torchbearer.callbacks.callbacks
        :members:
        :undoc-members:

Model Checkpointers
------------------------------------

..  automodule:: torchbearer.callbacks.checkpointers
        :members:
        :undoc-members:

Logging
------------------------------------

..  automodule:: torchbearer.callbacks.csv_logger
        :members:
        :undoc-members:

..  automodule:: torchbearer.callbacks.printer
        :members:
        :undoc-members:

..  automodule:: torchbearer.callbacks.timer
        :members:
        :undoc-members:

Tensorboard
------------------------------------

..  automodule:: torchbearer.callbacks.tensor_board
        :members:
        :undoc-members:

Early Stopping
------------------------------------

..  automodule:: torchbearer.callbacks.early_stopping
        :members:
        :undoc-members:

..  automodule:: torchbearer.callbacks.terminate_on_nan
        :members:
        :undoc-members:

Gradient Clipping
------------------------------------

..  automodule:: torchbearer.callbacks.gradient_clipping
        :members:
        :undoc-members:

Learning Rate Schedulers
------------------------------------

..  automodule:: torchbearer.callbacks.torch_scheduler
        :members:
        :undoc-members:

Weight Decay
------------------------------------

..  automodule:: torchbearer.callbacks.weight_decay
        :members:
        :undoc-members:

Decorators
------------------------------------

..  automodule:: torchbearer.callbacks.decorators
        :members:
        :undoc-members:
"""

from .callbacks import *
from .checkpointers import *
from .csv_logger import *
from .early_stopping import *
from .gradient_clipping import *
from .printer import *
from .tensor_board import *
from .terminate_on_nan import *
from .torch_scheduler import *
from .weight_decay import *
from .aggregate_predictions import *
from .decorators import *
from .timer import *
